package com.me.pacman.entity;

import com.badlogic.gdx.math.Vector2;
import com.me.pacman.entity.path.Path;
import com.me.pacman.level.LevelTile;
import com.me.pacman.state.LevelState;

public abstract class MovableEntity extends Entity {

    // speed in tiles / second.
    public float speed;
    public boolean moving;
    public Direction currDirection;
    private float tolerance;

    private Direction nextDirection = null;
    public boolean canMove = true;

    public Path currentPath = null;

    public MovableEntity(LevelState state, float x, float y, float speed, boolean moving, Direction currDirection, float turnTolerance) {
        super(state, x, y);
        this.speed = speed;
        this.moving = moving;
        this.currDirection = currDirection;
        this.tolerance = turnTolerance;
    }

    public void update(float dt) {
        super.update(dt);

        if (!moving) {
            return;
        }

        if (currentPath != null) {
            currentPath.update(dt);
            currentPath.updateEntity(this);
            return;
        }

        LevelTile nextTile = null;

        if (nextDirection != null) {
            boolean turned = false;
            nextTile = state.level.getTile(nextDirection.getVector().add(pos));
            if (nextTile == null) {
                // turned around immediately after existing tunnel (or trying to turn 90 degrees after entering tunnel)
                if (nextDirection.isOpposite(currDirection)) {
                    // only allow if turning 180 degrees
                    turned = true;
                }
            } else {
                if (nextTile.isPassable()) {
                    // it's possible to turn - are we close enough to our tile's center to make the turn?
                    Vector2 tileCenter = getTileVector(0.5f);
                    if (pos.dst2(tileCenter) <= tolerance * tolerance) {
                        // yes, move us to the center of our tile
                        pos = tileCenter;
                        turned = true;
                    }
                }
            }
            if (turned) {
                currDirection = nextDirection;
                nextDirection = null;
            }
        }

        LevelTile currentTile = state.level.getTile(pos);
        Vector2 new_pos = currDirection.getVector(speed * dt).add(pos);

        if (currentTile == null) {
            // handle traveling back onto the screen
            switch (currDirection) {
                case UP:
                    if (pos.y >= state.level.height + 1f) {
                        new_pos.y = -1f;
                    }
                    break;
                case DOWN:
                    if (pos.y <= 0) {
                        new_pos.y = state.level.height + 1f;
                    }
                    break;
                case LEFT:
                    if (pos.x <= 0) {
                        new_pos.x = state.level.width + 1f;
                    }
                    break;
                case RIGHT:
                    if (pos.x >= state.level.width + 1f) {
                        new_pos.x = -1f;
                    }
                    break;
            }
            canMove = true;
        } else {
            nextTile = state.level.getTile(currDirection.getVector(0.5f).add(new_pos));
            canMove = nextTile == null || nextTile.isPassable();
        }

        // if move isn't going to collide with wall, move normally.
        // otherwise, trim would-be decimal and center entity on tile
        pos.x = canMove? new_pos.x : ((int) new_pos.x) + 0.5f;
        pos.y = canMove? new_pos.y : ((int) new_pos.y) + 0.5f;
    }

    public void setNextDirection(Direction direction) {
        if (direction != currDirection) {
            nextDirection = direction;
        }
    }

    public Direction getNextDirection() {
        return nextDirection;
    }

}
